"""Module for managing xknx config parsing and validation."""
# flake8: noqa
from .config import Config, ConfigV1
from .schema import (
    BinarySensorSchema,
    ClimateSchema,
    ConnectionSchema,
    CoverSchema,
    DateTimeSchema,
    ExposeSchema,
    FanSchema,
    LightSchema,
    NotificationSchema,
    SceneSchema,
    SensorSchema,
    SwitchSchema,
    WeatherSchema,
    XKNXSchema,
)

__all__ = [
    "Config",
    "ConfigV1",
    "BinarySensorSchema",
    "ClimateSchema",
    "ConnectionSchema",
    "CoverSchema",
    "DateTimeSchema",
    "ExposeSchema",
    "FanSchema",
    "LightSchema",
    "NotificationSchema",
    "SceneSchema",
    "SensorSchema",
    "SwitchSchema",
    "WeatherSchema",
    "XKNXSchema",
]



"""Implementation of Basic KNX DPT B1U3 Values (DPT 3.007/3.008).

There are two separate dimming modes sharing the same DPT class:

 * Stepwise dimming
   The full brightness range is divided into 2^(stepcode-1) intervals.
   The value is always rounded to full interval boundary, i.e. 30% +25% = 50%, 50% +25% = 75%, 30% -25% = 25%

 * Start-stop dimming
   Dimming is started with -/+100% (0x1/0x9) and keeps dimming until a STOP diagram (0x0/0x8) is received.

As the same payload in these cases in interpreted completely different it is reasonable to make separate DPT classes.
"""
from enum import Enum
from typing import Tuple

from xknx.exceptions import ConversionError

from .dpt import DPTBase


class DPTControlStepCode(DPTBase):
    """Abstraction for KNX B1U3 values (DPT 3.007/3.008)."""

    # APCI (application layer control information)
    APCI_CONTROLMASK = 0x08
    APCI_STEPCODEMASK = 0x07
    APCI_MAX_VALUE = APCI_CONTROLMASK | APCI_STEPCODEMASK

    value_type = "control_stepcode"
    unit = ""
    payload_length = 1

    @classmethod
    def _encode(cls, control: bool, step_code: int):
        """Encode control-bit with step-code."""
        value = 1 if control > 0 else 0
        value = (value << 3) | (step_code & cls.APCI_STEPCODEMASK)
        return value

    @classmethod
    def _decode(cls, value) -> Tuple[bool, int]:
        """Decode value into control-bit and step-code."""
        control = 1 if (value & cls.APCI_CONTROLMASK) != 0 else 0
        step_code = value & cls.APCI_STEPCODEMASK
        return control, step_code

    @classmethod
    def _test_boundaries(cls, raw):
        """Test if raw KNX data is within defined range for this object."""
        if isinstance(raw, int):
            return 0 <= raw <= cls.APCI_MAX_VALUE
        return False

    @classmethod
    def _test_values(cls, control: bool, step_code: int):
        """Test if input values are valid."""
        if isinstance(control, int) and isinstance(step_code, int):
            if control in (0, 1) and 0 <= step_code <= cls.APCI_STEPCODEMASK:
                return True
        return False

    @classmethod
    def to_knx(cls, value, invert: bool = False):
        """Serialize to KNX/IP raw data."""
        if not isinstance(value, dict):
            raise ConversionError(
                "Cant serialize %s; invalid value type" % cls.__name__, value=value
            )

        try:
            control = value["control"]
            step_code = value["step_code"]
        except KeyError:
            raise ConversionError(
                "Cant serialize %s; invalid keys" % cls.__name__, value=value
            )

        if not cls._test_values(control, step_code):
            raise ConversionError(
                "Cant serialize %s; invalid values" % cls.__name__, value=value
            )

        if invert:
            control = 0 if control > 0 else 1

        return cls._encode(control, step_code)

    @classmethod
    def from_knx(cls, raw, invert: bool = False):
        """Parse/deserialize from KNX/IP raw data."""
        if not cls._test_boundaries(raw):
            raise ConversionError("Cant parse %s" % cls.__name__, raw=raw)

        control, step_code = cls._decode(raw)

        if invert:
            control = 0 if control > 0 else 1

        return {"control": control, "step_code": step_code}


class DPTControlStepwise(DPTControlStepCode):
    """Abstraction for KNX DPT 3.xxx in stepwise mode with conversion to an incement value."""

    dpt_main_number = 3
    dpt_sub_number = None
    value_type = "stepwise"
    unit = "%"

    @staticmethod
    def _from_increment(value):
        """Calculate control bit and stepcode as defined in the KNX standard section 3.3.1 from an increment value."""
        # control bit in KNX standard
        #   0: - = decrease/move up
        #   1: + = increase/move down
        control = 0 if value <= 0 else 1

        stepcode = (
            0  # special case = break indication (e.g. stop dimming/moving blinds)
        )
        if abs(value) >= 100:
            stepcode = 1
        elif abs(value) >= 50:
            stepcode = 2
        elif abs(value) >= 25:
            stepcode = 3
        elif abs(value) >= 12:
            stepcode = 4
        elif abs(value) >= 6:
            stepcode = 5
        elif abs(value) >= 3:
            stepcode = 6
        elif abs(value) >= 1:
            stepcode = 7

        return {"control": control, "step_code": stepcode}

    @staticmethod
    def _to_increment(value):
        """Calculate the increment value from the stepcode and control bit as defined in the KNX standard section 3.3.1."""
        # calculated using floor(100/2^((value&0x07)-1))
        inc = [0, 100, 50, 25, 12, 6, 3, 1][value["step_code"] & 0x07]
        return inc if value["control"] == 1 else -inc

    @classmethod
    def to_knx(cls, value, invert: bool = False):
        """Serialize to KNX/IP raw data."""
        if not isinstance(value, int):
            raise ConversionError("Cant serialize %s" % cls.__name__, value=value)

        return super().to_knx(cls._from_increment(value), invert)

    @classmethod
    def from_knx(cls, raw, invert: bool = False):
        """Parse/deserialize from KNX/IP raw data."""
        return cls._to_increment(super().from_knx(raw, invert))


class DPTControlStepwiseDimming(DPTControlStepwise):
    """Abstraction for KNX DPT 3.007 / DPT_Control_Dimming in stepwise mode."""

    dpt_main_number = 3
    dpt_sub_number = 7
    value_type = "stepwise_dimming"


class DPTControlStepwiseBlinds(DPTControlStepwise):
    """Abstraction for KNX DPT 3.008 / DPT_Control_Blinds in stepwise mode."""

    dpt_main_number = 3
    dpt_sub_number = 8
    value_type = "stepwise_blinds"


class TitleEnum(Enum):
    """Enum with a descriptive string representation.

    Ensures values are rendered nicely, e.g. in home assistant.
    """

    def __str__(self) -> str:
        """Return string representation."""
        # pylint: disable=no-member
        return self.name.title()


class DPTControlStartStop(DPTControlStepCode):
    """Abstraction for KNX DPT 3.xxx in start/stop mode."""

    value_type = "startstop"
    unit = ""

    class Direction(TitleEnum):
        """Enum for indicating the direction."""

        DECREASE = 0
        INCREASE = 1
        STOP = 2

    @classmethod
    def to_knx(cls, value, invert: bool = False):
        """Convert value to payload."""
        control = 0
        step_code = 0
        if value == cls.Direction(1):  # INCREASE/DOWN
            control = 1
            step_code = 1
        elif value == cls.Direction(0):  # DECREASE/UP
            control = 0
            step_code = 1
        elif value == cls.Direction(2):  # STOP
            control = 0
            step_code = 0
        else:
            raise ConversionError("Cant serialize %s" % cls.__name__, value=value)

        values = {"control": control, "step_code": step_code}
        return super().to_knx(values, invert)

    @classmethod
    def from_knx(cls, raw, invert: bool = False):
        """Convert current payload to value."""
        values = super().from_knx(raw, invert)
        if values["step_code"] == 0:
            return cls.Direction(2)  # STOP
        if values["control"] == 0:
            return cls.Direction(0)  # DECREASE/UP
        return cls.Direction(1)  # INCREASE/DOWN


class DPTControlStartStopDimming(DPTControlStartStop):
    """Abstraction for KNX DPT 3.007 / DPT_Control_Dimming in start/stop mode."""

    value_type = "startstop_dimming"

    # redefining Direction enum ensures proper typing, e.g.
    # DPTControlStartStop.Direction.INCREASE != DPTControlStartStopDimming.Direction.INCREASE
    class Direction(TitleEnum):
        """Enum for indicating the direction."""

        DECREASE = 0
        INCREASE = 1
        STOP = 2


class DPTControlStartStopBlinds(DPTControlStartStop):
    """Abstraction for KNX DPT 3.008 / DPT_Control_Blinds in start/stop mode."""

    value_type = "startstop_blinds"

    class Direction(TitleEnum):
        """Enum for indicating the direction."""

        UP = 0
        DOWN = 1
        STOP = 2



"""
Module for handling a vector/array of devices.

More or less an array with devices. Adds some search functionality to find devices.
"""
from typing import Awaitable, Callable, Iterator, List, Union

from xknx.telegram import GroupAddress, Telegram

from .device import Device

DeviceCallbackType = Callable[[Device], Awaitable[None]]


class Devices:
    """Class for handling a vector/array of devices."""

    def __init__(self) -> None:
        """Initialize Devices class."""
        self.__devices: List[Device] = []
        self.device_updated_cbs: List[DeviceCallbackType] = []

    def register_device_updated_cb(self, device_updated_cb: DeviceCallbackType) -> None:
        """Register callback for devices beeing updated."""
        self.device_updated_cbs.append(device_updated_cb)

    def unregister_device_updated_cb(
        self, device_updated_cb: DeviceCallbackType
    ) -> None:
        """Unregister callback for devices beeing updated."""
        self.device_updated_cbs.remove(device_updated_cb)

    def __iter__(self) -> Iterator[Device]:
        """Iterate registered devices."""
        yield from self.__devices

    def devices_by_group_address(self, group_address: GroupAddress) -> Iterator[Device]:
        """Return device(s) by group address."""
        for device in self.__devices:
            if device.has_group_address(group_address):
                yield device

    def __getitem__(self, key: Union[str, int]) -> Device:
        """Return device by name or by index."""
        for device in self.__devices:
            if device.name == key:
                return device
        if isinstance(key, int):
            return self.__devices[key]
        raise KeyError

    def __len__(self) -> int:
        """Return number of devices within vector."""
        return len(self.__devices)

    def __contains__(self, key: str) -> bool:
        """Return if devices with name 'key' is within devices."""
        for device in self.__devices:
            if device.name == key:
                return True
        return False

    def add(self, device: Device) -> None:
        """Add device to devices vector."""
        if not isinstance(device, Device):
            raise TypeError()
        device.register_device_updated_cb(self.device_updated)
        self.__devices.append(device)

    def remove(self, device: Device) -> None:
        """Remove device from devices vector."""
        self.__devices.remove(device)

    async def device_updated(self, device: Device) -> None:
        """Call all registered device updated callbacks of device."""
        for device_updated_cb in self.device_updated_cbs:
            await device_updated_cb(device)

    async def process(self, telegram: Telegram) -> None:
        """Process telegram."""
        if isinstance(telegram.destination_address, GroupAddress):
            for device in self.devices_by_group_address(telegram.destination_address):
                await device.process(telegram)

    async def sync(self) -> None:
        """Read state of devices from KNX bus."""
        for device in self.__devices:
            await device.sync()



"""Custom YAML loader."""
from collections import OrderedDict
import fnmatch
import logging
import os
from typing import Dict, Iterator, List, TypeVar, Union

from xknx.config.objects import NodeListClass, NodeStrClass
from xknx.exceptions import XKNXException
import yaml

logger = logging.getLogger("xknx.log")

JSON_TYPE = Union[List, Dict, str]  # pylint: disable=invalid-name
DICT_T = TypeVar("DICT_T", bound=Dict)  # pylint: disable=invalid-name


class SafeLineLoader(yaml.SafeLoader):
    """Loader class that keeps track of line numbers."""

    def compose_node(self, parent: yaml.nodes.Node, index: int) -> yaml.nodes.Node:
        """Annotate a node with the first line it was seen."""
        last_line: int = self.line
        node: yaml.nodes.Node = super().compose_node(parent, index)
        node.__line__ = last_line + 1  # type: ignore
        return node


def _add_reference(obj, loader: SafeLineLoader, node: yaml.nodes.Node):  # type: ignore
    """Add file reference information to an object."""
    if isinstance(obj, list):
        obj = NodeListClass(obj)
    if isinstance(obj, str):
        obj = NodeStrClass(obj)
    setattr(obj, "__config_file__", loader.name)
    setattr(obj, "__line__", node.start_mark.line)
    return obj


def load_yaml(fname: str) -> JSON_TYPE:
    """Load a YAML file."""
    try:
        with open(fname, encoding="utf-8") as conf_file:
            # If configuration file is empty YAML returns None
            # We convert that to an empty dict
            return yaml.load(conf_file, Loader=SafeLineLoader) or OrderedDict()
    except yaml.YAMLError as exc:
        logger.error(str(exc))
        raise XKNXException(exc) from exc
    except UnicodeDecodeError as exc:
        logger.error("Unable to read file %s: %s", fname, exc)
        raise XKNXException(exc) from exc


def _include_yaml(loader: SafeLineLoader, node: yaml.nodes.Node) -> JSON_TYPE:
    """Load another YAML file and embeds it using the !include tag.

    Example:
        device_tracker: !include device_tracker.yaml

    """
    fname = os.path.join(os.path.dirname(loader.name), node.value)
    try:
        return _add_reference(load_yaml(fname), loader, node)
    except FileNotFoundError as exc:
        raise XKNXException(f"{node.start_mark}: Unable to read file {fname}.") from exc


def _is_file_valid(name: str) -> bool:
    """Decide if a file is valid."""
    return not name.startswith(".")


def _find_files(directory: str, pattern: str) -> Iterator[str]:
    """Recursively load files in a directory."""
    for root, dirs, files in os.walk(directory, topdown=True):
        dirs[:] = [d for d in dirs if _is_file_valid(d)]
        for basename in sorted(files):
            if _is_file_valid(basename) and fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


def _construct_seq(loader: SafeLineLoader, node: yaml.nodes.Node) -> JSON_TYPE:
    """Add line number and file name to Load YAML sequence."""
    (obj,) = loader.construct_yaml_seq(node)
    return _add_reference(obj, loader, node)


def _include_dir_list_yaml(
    loader: SafeLineLoader, node: yaml.nodes.Node
) -> List[JSON_TYPE]:
    """Load multiple files from directory as a list."""
    loc = os.path.join(os.path.dirname(loader.name), node.value)
    return [load_yaml(f) for f in _find_files(loc, "*.yaml")]


def _env_var_yaml(loader: SafeLineLoader, node: yaml.nodes.Node) -> str:
    """Load environment variables and embed it into the configuration YAML."""
    args = node.value.split()

    # Check for a default value
    if len(args) > 1:
        return os.getenv(args[0], " ".join(args[1:]))
    if args[0] in os.environ:
        return os.environ[args[0]]
    logger.error("Environment variable %s not defined", node.value)
    raise XKNXException(node.value)


SafeLineLoader.add_constructor("!include", _include_yaml)
SafeLineLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, _construct_seq
)
SafeLineLoader.add_constructor("!env_var", _env_var_yaml)
SafeLineLoader.add_constructor("!include_dir_list", _include_dir_list_yaml)



"""
Module for managing an climate mode remote values.

DPT .
"""
from abc import abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
)

from xknx.dpt import (
    DPTArray,
    DPTBinary,
    DPTControllerStatus,
    DPTHVACContrMode,
    DPTHVACMode,
)
from xknx.dpt.dpt_hvac_mode import HVACControllerMode, HVACOperationMode
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import RemoteValue

if TYPE_CHECKING:
    from xknx.telegram.address import GroupAddressableType
    from xknx.xknx import XKNX

    AsyncCallback = Callable[[], Awaitable[None]]
    DPTPayload = Union[DPTArray, DPTBinary]

HVACModeType = TypeVar("HVACModeType", "HVACControllerMode", "HVACOperationMode")


class RemoteValueClimateModeBase(RemoteValue, Generic[HVACModeType]):
    """Base class for binary climate mode remote values."""

    @abstractmethod
    def supported_operation_modes(
        self,
    ) -> List["HVACModeType"]:
        """Return a list of all supported operation modes."""


class RemoteValueClimateMode(RemoteValueClimateModeBase[HVACModeType]):
    """Abstraction for remote value of KNX climate modes."""

    class ClimateModeType(Enum):
        """Implemented climate mode types."""

        CONTROLLER_STATUS = DPTControllerStatus
        HVAC_CONTR_MODE = DPTHVACContrMode
        HVAC_MODE = DPTHVACMode

    def __init__(
        self,
        xknx: "XKNX",
        group_address: Optional["GroupAddressableType"] = None,
        group_address_state: Optional["GroupAddressableType"] = None,
        sync_state: bool = True,
        device_name: Optional[str] = None,
        feature_name: str = "Climate Mode",
        climate_mode_type: Optional[ClimateModeType] = None,
        after_update_cb: Optional["AsyncCallback"] = None,
        passive_group_addresses: Optional[List["GroupAddressableType"]] = None,
    ):
        """Initialize remote value of KNX climate mode."""
        # pylint: disable=too-many-arguments
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
            passive_group_addresses=passive_group_addresses,
        )
        if not isinstance(climate_mode_type, self.ClimateModeType):
            raise ConversionError(
                "invalid climate mode type",
                climate_mode_type=str(climate_mode_type),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        self._climate_mode_transcoder = climate_mode_type.value

    def supported_operation_modes(self) -> List["HVACModeType"]:
        """Return a list of all supported operation modes."""
        return list(self._climate_mode_transcoder.SUPPORTED_MODES.values())

    @staticmethod
    def payload_valid(payload: Optional["DPTPayload"]) -> bool:
        """Test if telegram payload may be parsed."""
        return isinstance(payload, DPTArray) and len(payload.value) == 1

    def to_knx(self, value: Any) -> "DPTPayload":
        """Convert value to payload."""
        return DPTArray(self._climate_mode_transcoder.to_knx(value))

    def from_knx(self, payload: "DPTPayload") -> Optional[HVACModeType]:
        """Convert current payload to value."""
        # TODO: typing - remove cast
        return cast(
            Optional[HVACModeType],
            self._climate_mode_transcoder.from_knx(payload.value),
        )


class RemoteValueBinaryOperationMode(RemoteValueClimateModeBase[HVACOperationMode]):
    """Abstraction for remote value of split up KNX climate modes."""

    def __init__(
        self,
        xknx: "XKNX",
        group_address: Optional["GroupAddressableType"] = None,
        group_address_state: Optional["GroupAddressableType"] = None,
        sync_state: bool = True,
        device_name: Optional[str] = None,
        feature_name: str = "Climate Mode Binary",
        after_update_cb: Optional["AsyncCallback"] = None,
        operation_mode: Optional[HVACOperationMode] = None,
    ):
        """Initialize remote value of KNX DPT 1 representing a climate operation mode."""
        # pylint: disable=too-many-arguments
        if not isinstance(operation_mode, HVACOperationMode):
            raise ConversionError(
                "Invalid operation mode type",
                operation_mode=str(operation_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        if operation_mode not in self.supported_operation_modes():
            raise ConversionError(
                "Operation mode not supported for binary mode object",
                operation_mode=str(operation_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        self.operation_mode = operation_mode
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=True,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    @staticmethod
    def payload_valid(payload: Optional["DPTPayload"]) -> bool:
        """Test if telegram payload may be parsed."""
        return isinstance(payload, DPTBinary)

    def to_knx(self, value: Any) -> "DPTPayload":
        """Convert value to payload."""
        if isinstance(value, HVACOperationMode):
            # foreign operation modes will set the RemoteValue to False
            return DPTBinary(value == self.operation_mode)
        raise ConversionError(
            "value invalid",
            value=value,
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    @staticmethod
    def supported_operation_modes() -> List[HVACOperationMode]:
        """Return a list of the configured operation mode."""
        return [
            HVACOperationMode.COMFORT,
            HVACOperationMode.FROST_PROTECTION,
            HVACOperationMode.NIGHT,
            HVACOperationMode.STANDBY,
        ]

    def from_knx(self, payload: "DPTPayload") -> Optional[HVACOperationMode]:
        """Convert current payload to value."""
        if payload == DPTBinary(1):
            return self.operation_mode
        if payload == DPTBinary(0):
            return None
        raise CouldNotParseTelegram(
            "payload invalid",
            payload=str(payload),
            device_name=self.device_name,
            feature_name=self.feature_name,
        )


class RemoteValueBinaryHeatCool(RemoteValueClimateModeBase[HVACControllerMode]):
    """Abstraction for remote value of heat/cool controller mode."""

    def __init__(
        self,
        xknx: "XKNX",
        group_address: Optional["GroupAddressableType"] = None,
        group_address_state: Optional["GroupAddressableType"] = None,
        sync_state: bool = True,
        device_name: Optional[str] = None,
        feature_name: str = "Controller Mode Heat/Cool",
        after_update_cb: Optional["AsyncCallback"] = None,
        controller_mode: Optional[HVACControllerMode] = None,
    ):
        """Initialize remote value of KNX DPT 1 representing a climate controller mode."""
        # pylint: disable=too-many-arguments
        if not isinstance(controller_mode, HVACControllerMode):
            raise ConversionError(
                "Invalid controller mode type",
                controller_mode=str(controller_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        if controller_mode not in self.supported_operation_modes():
            raise ConversionError(
                "Controller mode not supported for binary mode object",
                controller_mode=str(controller_mode),
                device_name=str(device_name),
                feature_name=feature_name,
            )
        self.controller_mode = controller_mode
        super().__init__(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=True,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )

    @staticmethod
    def payload_valid(payload: Optional["DPTPayload"]) -> bool:
        """Test if telegram payload may be parsed."""
        return isinstance(payload, DPTBinary)

    @staticmethod
    def supported_operation_modes() -> List[HVACControllerMode]:
        """Return a list of the configured operation mode."""
        return [HVACControllerMode.HEAT, HVACControllerMode.COOL]

    def to_knx(self, value: Any) -> "DPTPayload":
        """Convert value to payload."""
        if isinstance(value, HVACControllerMode):
            # foreign operation modes will set the RemoteValue to False
            return DPTBinary(value == self.controller_mode)
        raise ConversionError(
            "value invalid",
            value=value,
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def from_knx(self, payload: "DPTPayload") -> Optional[HVACControllerMode]:
        """Convert current payload to value."""
        if payload == DPTBinary(1):
            return self.controller_mode
        if payload == DPTBinary(0):
            # return the other operation mode
            return next(
                (
                    _op
                    for _op in self.supported_operation_modes()
                    if _op is not self.controller_mode
                ),
                None,
            )
        raise CouldNotParseTelegram(
            "payload invalid",
            payload=str(payload),
            device_name=self.device_name,
            feature_name=self.feature_name,
        )



"""
Module for managing a remote value typically used within a sensor.

The module maps a given value_type to a DPT class and uses this class
for serialization and deserialization of the KNX value.
"""
from typing import List, Optional

from xknx.dpt import DPTArray, DPTBase
from xknx.exceptions import ConversionError

from .remote_value import RemoteValue


class RemoteValueSensor(RemoteValue):
    """Abstraction for many different sensor DPT types."""

    def __init__(
        self,
        xknx,
        group_address=None,
        group_address_state=None,
        sync_state=True,
        value_type=None,
        device_name=None,
        feature_name="Value",
        after_update_cb=None,
        passive_group_addresses: List[str] = None,
    ):
        """Initialize RemoteValueSensor class."""
        # pylint: disable=too-many-arguments
        _dpt_class = DPTBase.parse_transcoder(value_type)
        if _dpt_class is None:
            raise ConversionError(
                "invalid value type", value_type=value_type, device_name=device_name
            )
        self.dpt_class = _dpt_class
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
            passive_group_addresses=passive_group_addresses,
        )

    def payload_valid(self, payload):
        """Test if telegram payload may be parsed."""
        return (
            isinstance(payload, DPTArray)
            and len(payload.value) == self.dpt_class.payload_length
        )

    def to_knx(self, value):
        """Convert value to payload."""
        return DPTArray(self.dpt_class.to_knx(value))

    def from_knx(self, payload):
        """Convert current payload to value."""
        return self.dpt_class.from_knx(payload.value)

    @property
    def unit_of_measurement(self) -> Optional[str]:
        """Return the unit of measurement."""
        return self.dpt_class.unit

    @property
    def ha_device_class(self) -> Optional[str]:
        """Return a string representing the home assistant device class."""
        return getattr(self.dpt_class, "ha_device_class", None)



"""
Module for managing a control remote value.

Examples are switching commands with priority control, relative dimming or blinds control commands.
DPT 2.yyy and DPT 3.yyy
"""
from typing import List, Optional

from xknx.dpt import DPTBase, DPTBinary
from xknx.exceptions import ConversionError

from .remote_value import RemoteValue


class RemoteValueControl(RemoteValue):
    """Abstraction for remote value used for controling."""

    def __init__(
        self,
        xknx,
        group_address=None,
        group_address_state=None,
        sync_state=True,
        value_type=None,
        device_name=None,
        feature_name="Control",
        after_update_cb=None,
        invert=False,
        passive_group_addresses: List[str] = None,
    ):
        """Initialize control remote value."""
        # pylint: disable=too-many-arguments
        self.invert = invert
        _dpt_class = DPTBase.parse_transcoder(value_type)
        if _dpt_class is None:
            raise ConversionError(
                "invalid value type", value_type=value_type, device_name=device_name
            )
        self.dpt_class = _dpt_class
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
            passive_group_addresses=passive_group_addresses,
        )

    def payload_valid(self, payload):
        """Test if telegram payload may be parsed."""
        return isinstance(payload, DPTBinary)

    def to_knx(self, value):
        """Convert value to payload."""
        return DPTBinary(self.dpt_class.to_knx(value, invert=self.invert))

    def from_knx(self, payload):
        """Convert current payload to value."""
        return self.dpt_class.from_knx(payload.value, invert=self.invert)

    @property
    def unit_of_measurement(self) -> Optional[str]:
        """Return the unit of measurement."""
        return self.dpt_class.unit

    @property
    def ha_device_class(self) -> Optional[str]:
        """Return a string representing the home assistant device class."""
        return getattr(self.dpt_class, "ha_device_class", None)
